import argparse
import os
import sqlite3
import struct
import tempfile
from dataclasses import dataclass, field
from sqlite3 import Cursor

from acd.comments import CommentsRecord
from acd.comps import CompsRecord
from acd.dbextract import DbExtract
from acd.l5x.elements import Controller, ControllerBuilder
from acd.nameless import NamelessRecord
from acd.sbregion import SbRegionRecord
from acd.unzip import Unzip

from loguru import logger as log

@dataclass
class ExportL5x:
    input_filename: os.PathLike
    _temp_dir: str = "build" #tempfile.mkdtemp()
    _controller: Controller = None

    def __post_init__(self):
        log.info("Creating temporary directory (if it doesn't exist to store ACD database files - " + self._temp_dir)
        _DEFAULT_SQL_DATABASE_NAME = "acd.db"
        if os.path.exists(os.path.join(self._temp_dir, _DEFAULT_SQL_DATABASE_NAME)):
            os.remove(os.path.join(self._temp_dir, _DEFAULT_SQL_DATABASE_NAME))
        if not os.path.exists(os.path.join(self._temp_dir)):
            os.makedirs(self._temp_dir)
        log.info("Creating sqllite database to store ACD database records")
        self._db = sqlite3.connect(os.path.join(self._temp_dir, _DEFAULT_SQL_DATABASE_NAME))
        self._cur: Cursor = self._db.cursor()

        log.debug("Create Comps table in sqllite db")
        self._cur.execute("CREATE TABLE comps(object_id int, parent_id int, comp_name text, seq_number int, record_type int, record BLOB NOT NULL)")
        log.debug("Create pointers table in sqllite db")
        self._cur.execute("CREATE TABLE pointers(object_id int, parent_id int, comp_name text, seq_number int, record_type int, record BLOB NOT NULL)")
        log.debug("Create Rungs table in sqllite db")
        self._cur.execute("CREATE TABLE rungs(object_id int, rung text, seq_number int)")
        log.debug("Create Region_map table in sqllite db")
        self._cur.execute("CREATE TABLE region_map(object_id int, parent_id int, unknown int, seq_no int, record BLOB NOT NULL)")
        log.debug("Create Comments table in sqllite db")
        self._cur.execute(
            "CREATE TABLE comments(seq_number int, sub_record_length int, object_id int, record_string text, record_type int, parent int)")

        log.debug("Create Nameless table in sqllite db")
        self._cur.execute(
            "CREATE TABLE nameless(object_id int, parent_id int, record BLOB NOT NULL)")

        log.info("Extracting ACD database file")
        unzip = Unzip(self.input_filename)
        unzip.write_files(self._temp_dir)

        log.info("Getting records from ACD Comps file and storing in sqllite database")
        comps_db = DbExtract(os.path.join(self._temp_dir, "Comps.Dat")).read()
        for record in comps_db.records.record:
            CompsRecord(self._cur, record)
        self._db.commit()

        # Get a list of class ids for each collection
        # self._cur.execute("SELECT object_id, comp_name, record FROM comps WHERE parent_id=" + str(4240912631))
        # results = self._cur.fetchall()
        # classes = [[]]
        # for result in results:
        #     name = result[1]
        #     cip_class = hex(struct.unpack(
        #         "H", result[2][10: 10 + 2]
        #     )[0])
        #     classes.append([name, cip_class])


        log.info("Getting records from ACD Region Map file and storing in sqllite database")
        self.populate_region_map()

        log.info("Getting records from ACD SbRegion file and storing in sqllite database")
        sb_region_db = DbExtract(os.path.join(self._temp_dir, "SbRegion.Dat")).read()
        for record in sb_region_db.records.record:
            SbRegionRecord(self._cur, record)
        self._db.commit()

        log.info("Getting records from ACD Comments file and storing in sqllite database")
        comments_db = DbExtract(os.path.join(self._temp_dir, "Comments.Dat")).read()
        for record in comments_db.records.record:
            CommentsRecord(self._cur, record)
        self._db.commit()

        log.info("Getting records from ACD Nameless file and storing in sqllite database")
        nameless_db = DbExtract(os.path.join(self._temp_dir, "Nameless.Dat")).read()
        for record in nameless_db.records.record:
            NamelessRecord(self._cur, record)
        self._db.commit()

    @property
    def controller(self):
        if self._controller is None:
            self._controller = ControllerBuilder(self._cur).build()
        return self._controller


    def populate_region_map(self):
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=0 AND comp_name='Region Map'")
        results = self._cur.fetchall()

        if len(results) == 0:
            return
        record = results[0][3]

        identifier_offset = 70

        if len(record) < (identifier_offset + 8):
            return

        region_length = struct.unpack(
            "I", record[identifier_offset + 4: identifier_offset + 8]
        )[0]

        identifier_offset = 78
        record_length_absolute = identifier_offset + region_length - 4
        c = 0
        while identifier_offset <= (record_length_absolute - 16):
            parent_id_identifier = struct.unpack(
                "I", record[identifier_offset: identifier_offset + 4]
            )[0]

            unknown_identifier = struct.unpack(
                "I", record[identifier_offset + 4: identifier_offset + 8]
            )[0]

            seq_identifier = struct.unpack(
                "I", record[identifier_offset + 8: identifier_offset + 12]
            )[0]

            c += 1
            object_id_identifier = struct.unpack(
                "I", record[identifier_offset + 12: identifier_offset + 16]
            )[0]

            query: str = "INSERT INTO region_map VALUES (?, ?, ?, ?, ?)"
            enty: tuple = (object_id_identifier, parent_id_identifier, unknown_identifier, seq_identifier, record[identifier_offset: identifier_offset + 16])
            self._cur.execute(query, enty)
            identifier_offset += 16

            self._db.commit()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Read an ACD file and export the database as an L5X file')
    parser.add_argument('input', metavar='input', type=str, nargs='+',
                        help='The file to be converted')
    parser.add_argument('output', metavar='output', type=str, nargs='+',
                        help='Filename of the exported file')

    args = parser.parse_args()
    ExportL5x(args.input[0], args.output[0])


import os
from pathlib import Path

from acd.api import ExtractAcdDatabaseRecordsToFiles


def test_dump_database():
    database = ExtractAcdDatabaseRecordsToFiles(Path(os.path.join("..", "resources", "CuteLogix.ACD")), Path(os.path.join("build")))
    database.extract()



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class RxTag(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.header = RxTag.Header(self._io, self, self._root)
        self.parent_id = self._io.read_u4le()
        self.unique_tag_identifier = self._io.read_u4le()

    class Header(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.length = self._io.read_u4le()
            self.blank_1 = self._io.read_u4le()
            self.seq_number = self._io.read_u2le()
            self.record_type = self._io.read_u2le()
            self.blank_2 = self._io.read_u4le()
            self.object_id = self._io.read_u4le()
            self.parent_id = self._io.read_u4le()
            self.name = (self._io.read_bytes(124)).decode(u"UTF-16")


    @property
    def tag_name_length(self):
        if hasattr(self, '_m_tag_name_length'):
            return self._m_tag_name_length

        _pos = self._io.pos()
        self._io.seek(238)
        self._m_tag_name_length = self._io.read_u2le()
        self._io.seek(_pos)
        return getattr(self, '_m_tag_name_length', None)

    @property
    def data_type_id(self):
        if hasattr(self, '_m_data_type_id'):
            return self._m_data_type_id

        _pos = self._io.pos()
        self._io.seek(190)
        self._m_data_type_id = self._io.read_u4le()
        self._io.seek(_pos)
        return getattr(self, '_m_data_type_id', None)

    @property
    def tag_name(self):
        if hasattr(self, '_m_tag_name'):
            return self._m_tag_name

        _pos = self._io.pos()
        self._io.seek(240)
        self._m_tag_name = (self._io.read_bytes(self.tag_name_length)).decode(u"UTF-8")
        self._io.seek(_pos)
        return getattr(self, '_m_tag_name', None)

    @property
    def second_array_dimension(self):
        if hasattr(self, '_m_second_array_dimension'):
            return self._m_second_array_dimension

        _pos = self._io.pos()
        self._io.seek(178)
        self._m_second_array_dimension = self._io.read_u4le()
        self._io.seek(_pos)
        return getattr(self, '_m_second_array_dimension', None)

    @property
    def first_array_dimension(self):
        if hasattr(self, '_m_first_array_dimension'):
            return self._m_first_array_dimension

        _pos = self._io.pos()
        self._io.seek(174)
        self._m_first_array_dimension = self._io.read_u4le()
        self._io.seek(_pos)
        return getattr(self, '_m_first_array_dimension', None)

    @property
    def third_array_dimension(self):
        if hasattr(self, '_m_third_array_dimension'):
            return self._m_third_array_dimension

        _pos = self._io.pos()
        self._io.seek(182)
        self._m_third_array_dimension = self._io.read_u4le()
        self._io.seek(_pos)
        return getattr(self, '_m_third_array_dimension', None)





import pytest

from acd.unzip import Unzip


@pytest.fixture()
async def sample_acd():
    unzip = Unzip("../resources/CuteLogix.ACD")
    yield unzip


def test_open_file(sample_acd):
    assert sample_acd


def test_file_count(sample_acd):
    assert sample_acd.header.no_files == 25


def test_header_offset(sample_acd):
    assert sample_acd.header.record_offset == 2027550


def test_record_count(sample_acd):
    assert len(sample_acd.records) == 25


def test_filename(sample_acd):
    assert sample_acd.records[0].filename == "Version.Log"


def test_write_files(sample_acd):
    sample_acd.write_files("build")



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class FafaSbregions(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.record_length = self._io.read_u4le()
        self.header = FafaSbregions.Header(self._io, self, self._root)
        self.len_record_buffer = self._io.read_u4le()
        self.record_buffer = self._io.read_bytes(self.len_record_buffer)

    class Header(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.sb_regions = self._io.read_u2le()
            self.identifier = self._io.read_u4le()
            self.language_type = (KaitaiStream.bytes_terminate(self._io.read_bytes(41), 0, False)).decode(u"UTF-8")






import os
import shutil
import struct
from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from sqlite3 import Cursor
from typing import List

from acd.exceptions.CompsRecordException import UnknownRxTagVersion
from acd.generated.comps.rx_tag import RxTag


@dataclass
class L5xElementBuilder:
    _cur: Cursor
    _object_id: int = -1


@dataclass
class L5xElement:
    name: str


@dataclass
class DataType(L5xElement):
    children: List[str]


@dataclass
class Tag(L5xElement):
    data_table_instance: int
    data_type: str


@dataclass
class Routine(L5xElement):
    rungs: List[str]


@dataclass
class AOI(L5xElement):
    routines: List[Routine]
    tags: List[Tag]


@dataclass
class Program(L5xElement):
    routines: List[Routine]
    tags: List[Tag]


@dataclass
class Controller(L5xElement):
    data_types: List[DataType]
    tags: List[Tag]
    programs: List[Program]
    aois: List[AOI]


@dataclass
class DataTypeBuilder(L5xElementBuilder):

    def build(self) -> DataType:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        record = results[0][3]
        name = results[0][0]

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                self._object_id))
        member_results = self._cur.fetchall()
        if len(member_results) == 1:
            member_collection_id = member_results[0][1]

            self._cur.execute(
                "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                    member_collection_id))
            children_results = self._cur.fetchall()
            children = []
            for child in children_results:
                children.append(child[0])
            return DataType(name, children)
        return DataType(name, [])


@dataclass
class TagBuilder(L5xElementBuilder):

    def build(self) -> Tag:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        r = RxTag.from_bytes(results[0][3])

        if not r.body.valid:
            raise UnknownRxTagVersion(r.record_format_version)

        if r.body.data_type == 4294967295:
            data_type = ""
            name = r.body.name
        else:
            self._cur.execute(
                "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                    r.body.data_type))
            data_type_results = self._cur.fetchall()
            data_type = data_type_results[0][0]

            self._cur.execute(
                "SELECT seq_number, sub_record_length, object_id, record_string, record_type, parent FROM comments WHERE parent=" + str(
                    r.comment_id))
            comment_results = self._cur.fetchall()
            if r.body.tag_name_length == 16976:
                pass
            name = r.body.name
            if len(comment_results) > 0:
                pass
        if r.body.dimension_1 != 0:
            data_type = data_type + "[" + str(r.body.dimension_1) + "]"
        if r.body.dimension_2 != 0:
            data_type = data_type + "[" + str(r.body.dimension_2) + "]"
        if r.body.dimension_3 != 0:
            data_type = data_type + "[" + str(r.body.dimension_3) + "]"
        return Tag(name, r.body.data_table_instance, data_type)


@dataclass
class RoutineBuilder(L5xElementBuilder):

    def build(self) -> Routine:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        record = results[0][3]
        name = results[0][0]

        self._cur.execute(
            "SELECT object_id, parent_id, seq_no FROM region_map WHERE parent_id=" + str(
                self._object_id) +  " ORDER BY seq_no")
        results = self._cur.fetchall()
        rungs = []
        for member in results:
            self._cur.execute(
                "SELECT object_id, rung FROM rungs WHERE object_id=" + str(
                    member[0]))
            rungs_results = self._cur.fetchall()
            if len(rungs_results) > 0:
                rungs.append(rungs_results[0][1])
        return Routine(name, rungs)


@dataclass
class AoiBuilder(L5xElementBuilder):

    def build(self) -> AOI:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        record = results[0][3]
        name = results[0][0]

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxRoutineCollection'")
        collection_results = self._cur.fetchall()
        if len(collection_results) != 0:
            collection_id = collection_results[0][1]
        else:
            routines = []
            tags: List[Tag] = []
            return AOI(name, routines, tags)

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                collection_id))
        routine_results = self._cur.fetchall()

        routines = []
        for child in routine_results:
            routines.append(RoutineBuilder(self._cur, child[1]).build())

        # Get the Program Scoped Tags
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxTagCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one program tag collection")

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                results[0][1]))
        results = self._cur.fetchall()
        tags: List[Tag] = []
        for result in results:
            tags.append(TagBuilder(self._cur, result[1]).build())

        return AOI(name, routines, tags)


@dataclass
class ProgramBuilder(L5xElementBuilder):

    def build(self) -> Program:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE object_id=" + str(
                self._object_id))
        results = self._cur.fetchall()

        name = results[0][0]

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxRoutineCollection'")
        collection_results = self._cur.fetchall()
        collection_id = collection_results[0][1]

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record FROM comps WHERE parent_id=" + str(
                collection_id))
        routine_results = self._cur.fetchall()

        routines = []
        for child in routine_results:
            routines.append(RoutineBuilder(self._cur, child[1]).build())

        # Get the Program Scoped Tags
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxTagCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one program tag collection")

        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                results[0][1]))
        results = self._cur.fetchall()
        tags: List[Tag] = []
        for result in results:
            tags.append(TagBuilder(self._cur, result[1]).build())

        return Program(name, routines, tags)


@dataclass
class ControllerBuilder(L5xElementBuilder):

    def build(self) -> Controller:
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=0 AND record_type=256")
        results = self._cur.fetchall()
        if len(results) != 1:
            raise Exception("Does not contain exactly one root controller node")

        self._object_id = results[0][1]
        controller_name = results[0][0]

        # Get the data types
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxDataTypeCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one controller data type collection")

        _data_type_id = results[0][1]
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                _data_type_id))
        results = self._cur.fetchall()

        data_types: List[DataType] = []
        for result in results:
            _data_type_object_id = result[1]
            data_types.append(DataTypeBuilder(self._cur, _data_type_object_id).build())

        # Get the Controller Scoped Tags
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxTagCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one controller tag collection")
        _tag_collection_object_id = results[0][1]
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                _tag_collection_object_id))
        results = self._cur.fetchall()
        tags: List[Tag] = []
        for result in results:
            _tag_object_id = result[1]
            tags.append(TagBuilder(self._cur, _tag_object_id).build())

        # Get the Program Collection and get the programs
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxProgramCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one controller program collection")

        _program_collection_object_id = results[0][1]
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                _program_collection_object_id))
        results = self._cur.fetchall()
        programs: List[Program] = []
        for result in results:
            _program_object_id = result[1]
            programs.append(ProgramBuilder(self._cur, _program_object_id).build())

        # Get the AOI Collection and get the AOIs
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                self._object_id) + " AND comp_name='RxUDIDefinitionCollection'")
        results = self._cur.fetchall()
        if len(results) > 1:
            raise Exception("Contains more than one AOI collection")
        _aoi_collection_object_id = results[0][1]
        self._cur.execute(
            "SELECT comp_name, object_id, parent_id, record_type FROM comps WHERE parent_id=" + str(
                _aoi_collection_object_id))
        results = self._cur.fetchall()
        aois: List[AOI] = []
        for result in results:
            _aoi_object_id = result[1]
            aois.append(AoiBuilder(self._cur, _aoi_object_id).build())

        return Controller(controller_name, data_types, tags, programs, aois)

@dataclass
class DumpCompsRecords(L5xElementBuilder):
    base_directory: PathLike = Path("dump")

    def dump(self, parent_id: int = 0):
        self._cur.execute(
            f"SELECT comp_name, object_id, parent_id, record_type, record FROM comps WHERE parent_id={parent_id}")
        results = self._cur.fetchall()

        for result in results:
            object_id = result[1]
            name = result[0]
            record = result[4]
            new_path = Path(os.path.join(self.base_directory, name))
            if os.path.exists(os.path.join(new_path)):
                shutil.rmtree(os.path.join(new_path))
            if not os.path.exists(os.path.join(new_path)):
                os.makedirs(new_path)
            with open(Path(os.path.join(new_path, name + ".dat")), "wb") as file:
                file.write(record)

            DumpCompsRecords(self._cur, object_id, new_path).dump(object_id)





from typing import List

import pytest

from acd.comps import CompsRecord
from acd.dbextract import DbExtract
from acd.export_l5x import ExportL5x
from acd.sbregion import SbRegionRecord
from acd.unzip import Unzip

from loguru import logger as log


@pytest.fixture()
async def sample_acd():
    unzip = Unzip("../resources/CuteLogix.ACD").write_files("build")
    yield unzip


@pytest.fixture()
async def sbregion_dat():
    db = DbExtract("build/SbRegion.Dat")
    yield db


@pytest.fixture()
async def comps_dat():
    db = DbExtract("build/Comps.Dat").read()
    yield db 

@pytest.fixture(scope = "module")
def controller():
    log.level("DEBUG")
    yield ExportL5x("../resources/CuteLogix.ACD", "build").controller


def test_open_file(sample_acd, sbregion_dat):
    assert sbregion_dat


def test_parse_rungs_dat(controller):
    rung = controller.programs[-1].routines[-1].rungs[-1]
    assert rung == 'XIO(b_Timer[0].DN)TON(b_Timer[0],?,?);'


def test_parse_datatypes_dat(controller):
    data_type = controller.data_types[-1].name
    child =  controller.data_types[-1].children[-1]
    assert data_type == 'STRING20'
    assert child == 'DATA'


def test_parse_tags_dat(controller):
    tag_name = controller.tags[75].name
    data_type =  controller.tags[75].data_type
    assert data_type == 'BOOL'
    assert tag_name == 'Toggle'

def test_parse_comments_dat():
    db: DbExtract = DbExtract("build/Comments.Dat")


def test_parse_nameless_dat():
    db: DbExtract = DbExtract("build/Nameless.Dat")



import struct
from dataclasses import dataclass, InitVar
from io import BufferedReader
from pathlib import Path

from acd.generated.dat import Dat


@dataclass
class DatHeader:
    f: BufferedReader

    def __post_init__(self):
        self.start_position = self.f.seek(8)
        (
            self.total_length,
            self.region_pointer_offset,
            self._unknown1,
            self.no_records,
            self.no_records_table2,
        ) = struct.unpack("IIIII", self.f.read(20))
        self.f.seek(self.region_pointer_offset)
        if self.f.read(2) != b"\xfe\xfe":
            raise RuntimeError("Pointer Region Incorrect")
        (
            self.region_pointer_length,
            self.unknown2,
            self._unknown3,
            self.pointer_metadata_region,
            self.pointer_records_region,
        ) = struct.unpack("IIIII", self.f.read(20))
        self.f.seek(self.pointer_records_region)
        if self.f.read(2) != b"\xfe\xfe":
            raise RuntimeError("Record Region Incorrect")
        self.record_header_length, self._unknown4, self.unknown5, self.record_format = struct.unpack("IIII", self.f.read(16))
        if self.record_format == 132:
            raise RuntimeError("Cross Reference Databases Not Supported")
        elif self.record_format != 512:
            raise RuntimeError("Unknown record format")
        self.start_records_position = (
            self.pointer_records_region + self.record_header_length
        )
        self.f.seek(self.start_position)


@dataclass
class DatRecord:
    f: BufferedReader

    def __post_init__(self):
        self.identifier: bytes = self.f.read(2)
        self.record: bytes = bytearray()

        self.record_length = struct.unpack("I", self.f.read(4))[0]
        self.record: bytes = self.f.read(self.record_length - 6)


@dataclass
class DbExtract:
    filename: str

    def read(self) -> Dat:
        return Dat.from_file(self.filename)



import re
import struct
from dataclasses import dataclass
from sqlite3 import Cursor

from acd.dbextract import DatRecord
from acd.generated.sbregion.fafa_sbregions import FafaSbregions


@dataclass
class SbRegionRecord:
    _cur: Cursor
    dat_record: DatRecord

    def __post_init__(self):

        if self.dat_record.identifier == 64250:
            r = FafaSbregions.from_bytes(self.dat_record.record.record_buffer)
        else:
            return

        if r.header.language_type == "Rung NT":
            text = r.record_buffer.decode("utf-16-le").rstrip('\x00')
            self.text = self.replace_tag_references(text)

            query: str = "INSERT INTO rungs VALUES (?, ?, ?)"
            entry: tuple = (
                r.header.identifier, self.text, '')
            self._cur.execute(query, entry)
        elif r.header.language_type == "REGION AST":
            pass
        else:
            pass


    def replace_tag_references(self, sb_rec):
        m = re.findall("@[A-Za-z0-9]*@", sb_rec)
        for tag in m:
            tag_no = tag[1:-1]
            tag_id = int(tag_no, 16)
            self._cur.execute("SELECT object_id, comp_name FROM comps WHERE object_id=" + str(tag_id))
            results = self._cur.fetchall()
            sb_rec = sb_rec.replace(tag, results[0][1])
        return sb_rec



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Dat(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.header = Dat.Header(self._io, self, self._root)
        self._raw_records = self._io.read_bytes(((self.header.file_length - self.header.first_record_position) + 1))
        _io__raw_records = KaitaiStream(BytesIO(self._raw_records))
        self.records = Dat.Records(_io__raw_records, self, self._root)

    class FefeRecord(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.record_length = self._io.read_u4le()
            self.blank_1 = self._io.read_u4le()
            self.unknown_1 = self._io.read_u4le()
            self.unknown_2 = self._io.read_u4le()
            self.record_buffer = self._io.read_bytes(self.record_length)


    class FdfdRecord(KaitaiStruct):
        def __init__(self, record_length, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.record_length = record_length
            self._read()

        def _read(self):
            self.record_buffer = self._io.read_bytes_full()


    class BffbRecord(KaitaiStruct):
        def __init__(self, record_length, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.record_length = record_length
            self._read()

        def _read(self):
            self.record_buffer = self._io.read_bytes(self.record_length)


    class Header(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.format_type = self._io.read_u4le()
            self.blank_2 = self._io.read_u4le()
            self.file_length = self._io.read_u4le()
            self.first_record_position = self._io.read_u4le()
            self.blank_3 = self._io.read_u4le()
            self.number_records_fafa = self._io.read_u4le()
            self.header_buffer = []
            for i in range((self.first_record_position - 24)):
                self.header_buffer.append(self._io.read_u1())



    class Records(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.record = []
            i = 0
            while not self._io.is_eof():
                self.record.append(Dat.Record(self._io, self, self._root))
                i += 1



    class Record(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.identifier = self._io.read_u2le()
            if not  ((self.identifier == 65278) or (self.identifier == 65021) or (self.identifier == 64250) or (self.identifier == 64447)) :
                raise kaitaistruct.ValidationNotAnyOfError(self.identifier, self._io, u"/types/record/seq/0")
            self.record_length = self._io.read_u4le()
            _on = self.identifier
            if _on == 65278:
                self._raw_record = self._io.read_bytes((self.record_length - 6))
                _io__raw_record = KaitaiStream(BytesIO(self._raw_record))
                self.record = Dat.FefeRecord(_io__raw_record, self, self._root)
            elif _on == 64447:
                self._raw_record = self._io.read_bytes((self.record_length - 6))
                _io__raw_record = KaitaiStream(BytesIO(self._raw_record))
                self.record = Dat.BffbRecord((self.record_length - 6), _io__raw_record, self, self._root)
            elif _on == 64250:
                self._raw_record = self._io.read_bytes((self.record_length - 6))
                _io__raw_record = KaitaiStream(BytesIO(self._raw_record))
                self.record = Dat.FafaRecord((self.record_length - 6), _io__raw_record, self, self._root)
            elif _on == 65021:
                self._raw_record = self._io.read_bytes((self.record_length - 6))
                _io__raw_record = KaitaiStream(BytesIO(self._raw_record))
                self.record = Dat.FdfdRecord((self.record_length - 6), _io__raw_record, self, self._root)
            else:
                self.record = self._io.read_bytes((self.record_length - 6))


    class FafaRecord(KaitaiStruct):
        def __init__(self, record_length, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.record_length = record_length
            self._read()

        def _read(self):
            self.record_buffer = self._io.read_bytes(self.record_length)


    @property
    def third_array_dimension(self):
        if hasattr(self, '_m_third_array_dimension'):
            return self._m_third_array_dimension

        _pos = self._io.pos()
        self._io.seek(182)
        self._m_third_array_dimension = self._io.read_u4le()
        self._io.seek(_pos)
        return getattr(self, '_m_third_array_dimension', None)

    @property
    def data_type_id(self):
        if hasattr(self, '_m_data_type_id'):
            return self._m_data_type_id

        _pos = self._io.pos()
        self._io.seek(190)
        self._m_data_type_id = self._io.read_u4le()
        self._io.seek(_pos)
        return getattr(self, '_m_data_type_id', None)

    @property
    def tag_name_length(self):
        if hasattr(self, '_m_tag_name_length'):
            return self._m_tag_name_length

        _pos = self._io.pos()
        self._io.seek(238)
        self._m_tag_name_length = self._io.read_u2le()
        self._io.seek(_pos)
        return getattr(self, '_m_tag_name_length', None)

    @property
    def tag_name(self):
        if hasattr(self, '_m_tag_name'):
            return self._m_tag_name

        _pos = self._io.pos()
        self._io.seek(240)
        self._m_tag_name = (self._io.read_bytes(self.tag_name_length)).decode(u"UTF-8")
        self._io.seek(_pos)
        return getattr(self, '_m_tag_name', None)





import re
import struct
from dataclasses import dataclass
from sqlite3 import Cursor

from acd.dbextract import DatRecord
from acd.generated.comments.fafa_coments import FafaComents


@dataclass
class CommentsRecord:
    _cur: Cursor
    dat_record: DatRecord

    def __post_init__(self):
        if self.dat_record.identifier == 64250:
            r = FafaComents.from_bytes(self.dat_record.record.record_buffer)
            pass
        else:
            return

        query: str = "INSERT INTO comments VALUES (?, ?, ?, ?, ?, ?)"
        if r.header.record_type == 0x04:
            pass
        try:
            entry: tuple = (
                r.header.seq_number,
                r.header.sub_record_length,
                r.body.object_id,
                r.body.record_string,
                r.header.record_type,
                r.header.parent)
        except Exception as e:
            pass
        self._cur.execute(query, entry)

    def replace_tag_references(self, sb_rec):
        m = re.findall("@[A-Za-z0-9]*@", sb_rec)
        for tag in m:
            tag_no = tag[1:-1]
            tag_id = int(tag_no, 16)
            self._cur.execute("SELECT object_id, comp_name FROM comps WHERE object_id=" + str(tag_id))
            results = self._cur.fetchall()
            sb_rec = sb_rec.replace(tag, results[0][1])
        return sb_rec



import os
from pathlib import Path

from acd.api import ImportProjectFromFile, Project, Extract, ExtractAcdDatabase
from acd.export_l5x import ExportL5x


from acd.l5x.elements import DumpCompsRecords


def test_import_from_file():
    importer = ImportProjectFromFile(Path(os.path.join("..", "resources", "CuteLogix.ACD")))
    project: Project = importer.import_project()
    assert project is not None


def test_extract_database_files():
    extractor: Extract = ExtractAcdDatabase(Path(os.path.join("..", "resources", "CuteLogix.ACD")), Path(os.path.join("build")))
    extractor.extract()


def test_dump_to_files():
    export = ExportL5x(Path(os.path.join("..", "resources", "CuteLogix.ACD")))
    DumpCompsRecords(export._cur, 0).dump(0)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class FafaComps(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.record_length = self._io.read_u4le()
        self._raw_header = self._io.read_bytes(144)
        _io__raw_header = KaitaiStream(BytesIO(self._raw_header))
        self.header = FafaComps.Header(_io__raw_header, self, self._root)
        self.record_buffer = self._io.read_bytes(((self.record_length - 144) - 4))

    class Header(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def record_type(self):
            if hasattr(self, '_m_record_type'):
                return self._m_record_type

            _pos = self._io.pos()
            self._io.seek(6)
            self._m_record_type = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_record_type', None)

        @property
        def object_id(self):
            if hasattr(self, '_m_object_id'):
                return self._m_object_id

            _pos = self._io.pos()
            self._io.seek(12)
            self._m_object_id = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_object_id', None)

        @property
        def record_name(self):
            if hasattr(self, '_m_record_name'):
                return self._m_record_name

            _pos = self._io.pos()
            self._io.seek(20)
            self._raw__m_record_name = self._io.read_bytes(124)
            _io__raw__m_record_name = KaitaiStream(BytesIO(self._raw__m_record_name))
            self._m_record_name = FafaComps.StrzUtf16(_io__raw__m_record_name, self, self._root)
            self._io.seek(_pos)
            return getattr(self, '_m_record_name', None)

        @property
        def seq_number(self):
            if hasattr(self, '_m_seq_number'):
                return self._m_seq_number

            _pos = self._io.pos()
            self._io.seek(4)
            self._m_seq_number = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_seq_number', None)

        @property
        def parent_id(self):
            if hasattr(self, '_m_parent_id'):
                return self._m_parent_id

            _pos = self._io.pos()
            self._io.seek(16)
            self._m_parent_id = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_parent_id', None)


    class StrzUtf16(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.value = (self._io.read_bytes((2 * (len(self.code_units) - 1)))).decode(u"utf-16le")
            self.term = self._io.read_u2le()
            if not self.term == 0:
                raise kaitaistruct.ValidationNotEqualError(0, self.term, self._io, u"/types/strz_utf_16/seq/1")

        @property
        def code_units(self):
            if hasattr(self, '_m_code_units'):
                return self._m_code_units

            _pos = self._io.pos()
            self._io.seek(self._io.pos())
            self._m_code_units = []
            i = 0
            while True:
                _ = self._io.read_u2le()
                self._m_code_units.append(_)
                if _ == 0:
                    break
                i += 1
            self._io.seek(_pos)
            return getattr(self, '_m_code_units', None)






import re
import struct
from dataclasses import dataclass
from sqlite3 import Cursor

from acd.dbextract import DatRecord


@dataclass
class NamelessRecord:
    _cur: Cursor
    dat_record: DatRecord

    def __post_init__(self):
        if self.dat_record.identifier == b'\xfa\xfa':
            identifier_offset = 8
            self.identifier = struct.unpack(
                "I", self.dat_record.record[identifier_offset : identifier_offset + 4]
            )[0]

            not_sure_offset = 12
            self.not_sure = struct.unpack(
                "I", self.dat_record.record[not_sure_offset: not_sure_offset + 4]
            )[0]

            query: str = "INSERT INTO nameless VALUES (?, ?, ?)"
            enty: tuple = (self.identifier, self.not_sure, self.dat_record.record)
            self._cur.execute(query, enty)



import struct
from dataclasses import dataclass
from io import BytesIO
from sqlite3 import Cursor

from kaitaistruct import KaitaiStream

from acd.dbextract import DatRecord
from acd.generated.comps.fafa_comps import FafaComps
from acd.generated.comps.fdfd_comps import FdfdComps


@dataclass
class RecordData:
    object_id: int
    record_length: int
    seq_number: int
    record_type: int
    dat_record: DatRecord


@dataclass
class CompsRecord:
    _cur: Cursor
    dat_record: DatRecord

    def __post_init__(self):

        if self.dat_record.identifier == 64250:
            r = FafaComps.from_bytes(self.dat_record.record.record_buffer)
        elif self.dat_record.identifier == 65021:
            r = FdfdComps(self.dat_record.record_length, KaitaiStream(BytesIO(self.dat_record.record.record_buffer)))
        else:
            return

        query: str = f"DELETE FROM comps WHERE object_id={r.header.object_id}"
        self._cur.execute(query)

        query: str = "INSERT INTO comps VALUES (?, ?, ?, ?, ?, ?)"
        ss = r.header.record_name.value
        entry: tuple = (
            r.header.object_id, r.header.parent_id, r.header.record_name.value, r.header.seq_number, r.header.record_type,
            r.record_buffer)
        self._cur.execute(query, entry)



from dataclasses import dataclass
from os import PathLike

from acd.database.acd_database import AcdDatabase
from acd.l5x.elements import Controller, DumpCompsRecords

from acd.export_l5x import ExportL5x
from acd.unzip import Unzip


# Returned Project Structures
@dataclass
class Project:
    """Controller Project"""
    controller: Controller


# Import Export Interfaces
class ImportProject:
    """"Interface to import an PLC project"""

    def import_project(self) -> Project:
        # Import Project Interface
        pass


class ExportProject:
    """"Interface to export an PLC project"""

    def export_project(self, project: Project):
        # Export Project Interface
        pass


# Concreate examples of importing and exporting projects
@dataclass
class ImportProjectFromFile(ImportProject):
    """Import a Controller from an ACD stored on file"""
    filename: PathLike

    def import_project(self) -> Project:
        # Import Project Interface
        export = ExportL5x(self.filename)
        return Project(export.controller)


@dataclass
class ExportProjectToFile(ExportProject):
    """Export a Controller to an ACD file"""
    filename: PathLike

    def export_project(self, project: Project):
        # Concreate example of exporting a Project Object to an ACD file
        raise NotImplementedError


# Extracting/Compressing files from an ACD file Interfaces
class Extract:
    """Base class for all extract functions"""

    def extract(self):
        # Interface for extracting database files
        pass


class Compress:
    """Base class for all compress functions"""

    def compress(self):
        # Interface for extracting database files
        pass


# Concreate examples of extracting and compressing ACD files
@dataclass
class ExtractAcdDatabase(Extract):
    """Extract database files from a Logix ACD file"""
    filename: PathLike
    output_directory: PathLike

    def extract(self):
        # Implement the extraction of an ACD file
        unzip = Unzip(self.filename)
        unzip.write_files(self.output_directory)


@dataclass
class CompressAcdDatabase(Extract):
    """Compress database files to a Logix ACD file"""
    filename: PathLike
    output_directory: PathLike

    def compress(self):
        # Implement the compressing of an ACD file
        raise NotImplementedError


@dataclass
class ExtractAcdDatabaseRecordsToFiles(ExportProject):
    """Export all ACD databases to a raw database record tree"""
    filename: PathLike
    output_directory: PathLike

    def extract(self):
        # Implement the extraction of an ACD file
        database = AcdDatabase(self.filename, self.output_directory)
        database.extract_to_file()


@dataclass
class DumpCompsRecordsToFile(ExportProject):
    """
    Dump the Comps database to a folder. Each individual record can then be navigated and viewed.

    :param str filename: Filename of ACD file
    :param str output_directory: Location to store the records
    """
    filename: PathLike
    output_directory: PathLike

    def extract(self):
        export = ExportL5x(self.filename)
        DumpCompsRecords(export._cur, 0).dump(0)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import subprocess

from setuptools import setup, find_packages
from setuptools.command.install import install as _install

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

class install(_install):
    def run(self):
        subprocess.run(["ksc", "-t", "python", "--outdir", "acd/generated/", "--python-package", "acd.generated", "resources/templates/Dat/Dat.ksy"])
        subprocess.run(["ksc", "-t", "python", "--outdir", "acd/generated/comps/", "--python-package", "acd.generated.comps", "resources/templates/Comps/RxTag.ksy"])
        subprocess.run(
            ["ksc", "-t", "python", "--outdir", "acd/generated/comps/", "--python-package", "acd.generated.comps",
             "resources/templates/Comps/FAFA_Comps.ksy"])
        subprocess.run(
            ["ksc", "-t", "python", "--outdir", "acd/generated/sbregion/", "--python-package", "acd.generated.sbregion",
             "resources/templates/SbRegion/FAFA_SbRegion.ksy"])
        subprocess.run(
            ["ksc", "-t", "python", "--outdir", "acd/generated/comps/", "--python-package", "acd.generated.comps",
             "resources/templates/Comps/FDFD_Comps.ksy"])
        subprocess.run(
            ["ksc", "-t", "python", "--outdir", "acd/generated/comments/", "--python-package", "acd.generated.comments",
             "resources/templates/Comments/FAFA_Comments.ksy"])
        _install.run(self)
        print("--------------------------------------------------------------------")

setup(
    name="acd-tools",
    version="0.1a6",
    description="Rockwell ACD File Tools",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    ],
    keywords="rockwell acd logix",
    url="https://github.com/hutcheb/acd",
    author="Ben Hutcheson",
    author_email="",
    license="Apache 2.0",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(include=["acd", "acd.*"]),
    setup_requires=[
        "wheel",
    ],
    install_requires=[
        "loguru",
        "kaitaistruct",
    ],
    extras_require={
        "dev": [
            "requires",
            "pytest-asyncio>=0.18.3",
            "pip-tools",
            "pre-commit>=2.6.0",
            "pytest-mock>=3.3.1",
            "mock>=4.0.2",
            "mypy>=0.942",
            "flake8>=4.0.1",
        ]
    },
    cmdclass={'install': install},
)



import gzip
import os
import struct
from dataclasses import dataclass, InitVar
from io import BufferedReader
from pathlib import Path


@dataclass
class AcdHeader:
    f: BufferedReader

    def __post_init__(self):
        self.file_size = self.f.seek(-8, 2) + 8
        self.no_files, self._unknown_two = struct.unpack("II", self.f.read(8))
        self.record_offset = self.file_size - self.no_files * 528 - 8
        self.f.seek(0, 0)


@dataclass
class FileRecord:
    f: BufferedReader

    def __post_init__(self):
        self._read_filename()
        self.file_length, self.file_offset = struct.unpack("II", self.f.read(8))

    def _read_filename(self):
        end_found = False
        byte_array = b""
        while not end_found:
            # Keep reading until a null character is found
            b = self.f.read(2)
            if b == b"\x00\x00":
                end_found = True
            else:
                byte_array += b
        filename_length = len(byte_array)
        self.f.seek(520 - filename_length - 2, 1)
        self.filename = byte_array.decode("utf-16-le")


@dataclass
class Unzip:
    filename: InitVar[os.PathLike]

    def __post_init__(self, filename):
        self._filename = Path(filename)
        self._read()

    def write_files(self, directory: os.PathLike):
        Path(directory).mkdir(parents=True, exist_ok=True)
        with open(self._filename, "rb") as f:
            for record in self.records:
                f.seek(record.file_offset)
                with open(os.path.join(directory, record.filename), "wb") as out_file:
                    if f.read(2) == b"\x1f\x8b":
                        # Uncompress the files with gzip compression, the magic number is handy here.
                        f.seek(record.file_offset)
                        out_file.write(gzip.decompress(f.read(record.file_length)))
                    else:
                        f.seek(record.file_offset)
                        out_file.write(f.read(record.file_length))

    def _read_magic_number(self, f: BufferedReader):
        # Magic number is just the start of the first file in the archive
        # Even so it is a handy indicator
        magic_bytes = f.read(2)
        if magic_bytes != b"\x0d\x0a":
            raise RuntimeError("File isn't a Rockwell ACD file")
        f.seek(0, 0)

    def _read_file_header(self, f: BufferedReader):
        self.header: AcdHeader = AcdHeader(f)

    def _read_records(self, f: BufferedReader):
        f.seek(self.header.record_offset)
        self.records = []
        for _ in range(0, self.header.no_files):
            self.records.append(FileRecord(f))

    def _read(self):
        with open(self._filename, "rb") as f:
            # Check the file's magic number to confirm an ACD file
            self._read_magic_number(f)
            self._read_file_header(f)
            self._read_records(f)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO


if getattr(kaitaistruct, 'API_VERSION', (0, 9)) < (0, 9):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class RxTag(KaitaiStruct):
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.parent_id = self._io.read_u4le()
        self.unique_tag_identifier = self._io.read_u4le()
        self.record_format_version = self._io.read_u2le()
        self.comment_id = self._io.read_u4le()
        _on = self.record_format_version
        if _on == 60:
            self.body = RxTag.V60(self._io, self, self._root)
        elif _on == 63:
            self.body = RxTag.V63(self._io, self, self._root)
        else:
            self.body = RxTag.VUnknown(self._io, self, self._root)

    class V63(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def cip_data_type(self):
            if hasattr(self, '_m_cip_data_type'):
                return self._m_cip_data_type

            _pos = self._io.pos()
            self._io.seek(66)
            self._m_cip_data_type = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_cip_data_type', None)

        @property
        def tag_name_length(self):
            if hasattr(self, '_m_tag_name_length'):
                return self._m_tag_name_length

            _pos = self._io.pos()
            self._io.seek((78 + (len(self.records) * 4)))
            self._m_tag_name_length = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_tag_name_length', None)

        @property
        def data_type(self):
            if hasattr(self, '_m_data_type'):
                return self._m_data_type

            _pos = self._io.pos()
            self._io.seek(42)
            self._m_data_type = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_data_type', None)

        @property
        def dimension_2(self):
            if hasattr(self, '_m_dimension_2'):
                return self._m_dimension_2

            _pos = self._io.pos()
            self._io.seek(30)
            self._m_dimension_2 = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_dimension_2', None)

        @property
        def dimension_3(self):
            if hasattr(self, '_m_dimension_3'):
                return self._m_dimension_3

            _pos = self._io.pos()
            self._io.seek(34)
            self._m_dimension_3 = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_dimension_3', None)

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = True
            return getattr(self, '_m_valid', None)

        @property
        def sub_record_length(self):
            if hasattr(self, '_m_sub_record_length'):
                return self._m_sub_record_length

            _pos = self._io.pos()
            self._io.seek((78 + (len(self.records) * 4)))
            self._m_sub_record_length = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_sub_record_length', None)

        @property
        def logical_path(self):
            if hasattr(self, '_m_logical_path'):
                return self._m_logical_path

            _pos = self._io.pos()
            self._io.seek(666)
            self._m_logical_path = RxTag.LogicalPath(self._io, self, self._root)
            self._io.seek(_pos)
            return getattr(self, '_m_logical_path', None)

        @property
        def name(self):
            if hasattr(self, '_m_name'):
                return self._m_name

            _pos = self._io.pos()
            self._io.seek(((78 + (len(self.records) * 4)) + 2))
            self._m_name = (self._io.read_bytes(self.tag_name_length)).decode(u"UTF-8")
            self._io.seek(_pos)
            return getattr(self, '_m_name', None)

        @property
        def dimension_1(self):
            if hasattr(self, '_m_dimension_1'):
                return self._m_dimension_1

            _pos = self._io.pos()
            self._io.seek(26)
            self._m_dimension_1 = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_dimension_1', None)

        @property
        def records(self):
            if hasattr(self, '_m_records'):
                return self._m_records

            _pos = self._io.pos()
            self._io.seek(78)
            self._m_records = []
            i = 0
            while True:
                _ = self._io.read_u4le()
                self._m_records.append(_)
                if _ == 590:
                    break
                i += 1
            self._io.seek(_pos)
            return getattr(self, '_m_records', None)

        @property
        def data_table_instance(self):
            if hasattr(self, '_m_data_table_instance'):
                return self._m_data_table_instance

            _pos = self._io.pos()
            self._io.seek(50)
            self._m_data_table_instance = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_data_table_instance', None)


    class V60(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def cip_data_type(self):
            if hasattr(self, '_m_cip_data_type'):
                return self._m_cip_data_type

            _pos = self._io.pos()
            self._io.seek(66)
            self._m_cip_data_type = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_cip_data_type', None)

        @property
        def tag_name_length(self):
            if hasattr(self, '_m_tag_name_length'):
                return self._m_tag_name_length

            _pos = self._io.pos()
            self._io.seek(90)
            self._m_tag_name_length = self._io.read_u2le()
            self._io.seek(_pos)
            return getattr(self, '_m_tag_name_length', None)

        @property
        def data_type(self):
            if hasattr(self, '_m_data_type'):
                return self._m_data_type

            _pos = self._io.pos()
            self._io.seek(42)
            self._m_data_type = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_data_type', None)

        @property
        def dimension_2(self):
            if hasattr(self, '_m_dimension_2'):
                return self._m_dimension_2

            _pos = self._io.pos()
            self._io.seek(30)
            self._m_dimension_2 = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_dimension_2', None)

        @property
        def dimension_3(self):
            if hasattr(self, '_m_dimension_3'):
                return self._m_dimension_3

            _pos = self._io.pos()
            self._io.seek(34)
            self._m_dimension_3 = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_dimension_3', None)

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = True
            return getattr(self, '_m_valid', None)

        @property
        def logical_path(self):
            if hasattr(self, '_m_logical_path'):
                return self._m_logical_path

            _pos = self._io.pos()
            self._io.seek(666)
            self._m_logical_path = RxTag.LogicalPath(self._io, self, self._root)
            self._io.seek(_pos)
            return getattr(self, '_m_logical_path', None)

        @property
        def name(self):
            if hasattr(self, '_m_name'):
                return self._m_name

            _pos = self._io.pos()
            self._io.seek(92)
            self._m_name = (self._io.read_bytes(self.tag_name_length)).decode(u"UTF-8")
            self._io.seek(_pos)
            return getattr(self, '_m_name', None)

        @property
        def dimension_1(self):
            if hasattr(self, '_m_dimension_1'):
                return self._m_dimension_1

            _pos = self._io.pos()
            self._io.seek(26)
            self._m_dimension_1 = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_dimension_1', None)

        @property
        def data_table_instance(self):
            if hasattr(self, '_m_data_table_instance'):
                return self._m_data_table_instance

            _pos = self._io.pos()
            self._io.seek(50)
            self._m_data_table_instance = self._io.read_u4le()
            self._io.seek(_pos)
            return getattr(self, '_m_data_table_instance', None)


    class LogicalPath(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.position_0 = self._io.read_u4le()
            self.position_1 = self._io.read_u4le()
            self.position_2 = self._io.read_u4le()
            self.position_3 = self._io.read_u4le()
            self.position_4 = self._io.read_u4le()


    class VUnknown(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def valid(self):
            if hasattr(self, '_m_valid'):
                return self._m_valid

            self._m_valid = False
            return getattr(self, '_m_valid', None)


    class V63Records(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.unknown_0 = self._io.read_u4le()
            self.unknown_1 = self._io.read_u4le()
            self.unknown_2 = self._io.read_u4le()



